package java.awt;

import java.util.WeakHashMap;
import sun.awt.AWTAccessor;
import java.awt.peer.LightweightPeer;
import java.awt.event.WindowEvent;
import java.awt.event.FocusEvent;
import sun.awt.SunToolkit;
import java.awt.event.KeyEvent;
import java.util.List;
import java.beans.VetoableChangeListener;
import java.beans.PropertyChangeListener;
import java.util.Iterator;
import java.security.AccessController;
import java.security.PrivilegedAction;
import java.beans.PropertyVetoException;
import sun.awt.KeyboardFocusManagerPeerProvider;
import java.util.HashSet;
import java.util.Collections;
import java.util.StringTokenizer;
import sun.awt.AppContext;
import java.lang.reflect.Field;
import java.lang.ref.WeakReference;
import java.util.Map;
import java.util.LinkedList;
import java.beans.PropertyChangeSupport;
import java.beans.VetoableChangeSupport;
import java.util.Set;
import java.awt.peer.KeyboardFocusManagerPeer;
import sun.util.logging.PlatformLogger;

public abstract class KeyboardFocusManager implements KeyEventDispatcher, KeyEventPostProcessor
{
    private static final PlatformLogger focusLog;
    transient KeyboardFocusManagerPeer peer;
    private static final PlatformLogger log;
    public static final int FORWARD_TRAVERSAL_KEYS = 0;
    public static final int BACKWARD_TRAVERSAL_KEYS = 1;
    public static final int UP_CYCLE_TRAVERSAL_KEYS = 2;
    public static final int DOWN_CYCLE_TRAVERSAL_KEYS = 3;
    static final int TRAVERSAL_KEY_LENGTH = 4;
    private static Component focusOwner;
    private static Component permanentFocusOwner;
    private static Window focusedWindow;
    private static Window activeWindow;
    private FocusTraversalPolicy defaultPolicy;
    private static final String[] defaultFocusTraversalKeyPropertyNames;
    private static final AWTKeyStroke[][] defaultFocusTraversalKeyStrokes;
    private Set<AWTKeyStroke>[] defaultFocusTraversalKeys;
    private static Container currentFocusCycleRoot;
    private VetoableChangeSupport vetoableSupport;
    private PropertyChangeSupport changeSupport;
    private LinkedList<KeyEventDispatcher> keyEventDispatchers;
    private LinkedList<KeyEventPostProcessor> keyEventPostProcessors;
    private static Map<Window, WeakReference<Component>> mostRecentFocusOwners;
    private static AWTPermission replaceKeyboardFocusManagerPermission;
    transient SequencedEvent currentSequencedEvent;
    private static LinkedList<KeyboardFocusManager.HeavyweightFocusRequest> heavyweightRequests;
    private static LinkedList<KeyboardFocusManager.LightweightFocusRequest> currentLightweightRequests;
    private static boolean clearingCurrentLightweightRequests;
    private static boolean allowSyncFocusRequests;
    private static Component newFocusOwner;
    private static volatile boolean disableRestoreFocus;
    static final int SNFH_FAILURE = 0;
    static final int SNFH_SUCCESS_HANDLED = 1;
    static final int SNFH_SUCCESS_PROCEED = 2;
    static Field proxyActive;
    
    private static native void initIDs();
    
    public static KeyboardFocusManager getCurrentKeyboardFocusManager() {
        return getCurrentKeyboardFocusManager(AppContext.getAppContext());
    }
    
    static synchronized KeyboardFocusManager getCurrentKeyboardFocusManager(AppContext appContext) {
        KeyboardFocusManager value = (KeyboardFocusManager)appContext.get(KeyboardFocusManager.class);
        if (value == null) {
            value = new DefaultKeyboardFocusManager();
            appContext.put(KeyboardFocusManager.class, value);
            return value;
        }
        return value;
    }
    
    public static void setCurrentKeyboardFocusManager(KeyboardFocusManager value) throws SecurityException {
        checkReplaceKFMPermission();
        KeyboardFocusManager keyboardFocusManager = null;
        synchronized (KeyboardFocusManager.class) {
            AppContext appContext = AppContext.getAppContext();
            if (value != null) {
                keyboardFocusManager = getCurrentKeyboardFocusManager(appContext);
                appContext.put(KeyboardFocusManager.class, value);
            }
            else {
                keyboardFocusManager = getCurrentKeyboardFocusManager(appContext);
                appContext.remove(KeyboardFocusManager.class);
            }
        }
        if (keyboardFocusManager != null) {
            keyboardFocusManager.firePropertyChange("managingFocus", Boolean.TRUE, Boolean.FALSE);
        }
        if (value != null) {
            value.firePropertyChange("managingFocus", Boolean.FALSE, Boolean.TRUE);
        }
    }
    
    final void setCurrentSequencedEvent(SequencedEvent currentSequencedEvent) {
        synchronized (SequencedEvent.class) {
            assert this.currentSequencedEvent == null;
            this.currentSequencedEvent = currentSequencedEvent;
        }
    }
    
    final SequencedEvent getCurrentSequencedEvent() {
        synchronized (SequencedEvent.class) {
            return this.currentSequencedEvent;
        }
    }
    
    static Set<AWTKeyStroke> initFocusTraversalKeysSet(String str, Set<AWTKeyStroke> s) {
        StringTokenizer stringTokenizer = new StringTokenizer(str, ",");
        while (stringTokenizer.hasMoreTokens()) {
            s.add(AWTKeyStroke.getAWTKeyStroke(stringTokenizer.nextToken()));
        }
        return s.isEmpty() ? Collections.EMPTY_SET : Collections.unmodifiableSet(s);
    }
    
    public KeyboardFocusManager() {
        this.defaultPolicy = new DefaultFocusTraversalPolicy();
        this.defaultFocusTraversalKeys = new Set[4];
        this.currentSequencedEvent = null;
        Object s;
        for (int i = 0; i < 4; this.defaultFocusTraversalKeys[i] = (s.isEmpty() ? Collections.EMPTY_SET : Collections.unmodifiableSet(s)), ++i) {
            s = new HashSet();
            for (int j = 0; j < KeyboardFocusManager.defaultFocusTraversalKeyStrokes[i].length; ++j) {
                s.add(KeyboardFocusManager.defaultFocusTraversalKeyStrokes[i][j]);
            }
        }
        this.initPeer();
    }
    
    private void initPeer() {
        this.peer = ((KeyboardFocusManagerPeerProvider)Toolkit.getDefaultToolkit()).getKeyboardFocusManagerPeer();
    }
    
    public Component getFocusOwner() {
        synchronized (KeyboardFocusManager.class) {
            if (KeyboardFocusManager.focusOwner == null) {
                return null;
            }
            return (KeyboardFocusManager.focusOwner.appContext == AppContext.getAppContext()) ? KeyboardFocusManager.focusOwner : null;
        }
    }
    
    protected Component getGlobalFocusOwner() throws SecurityException {
        synchronized (KeyboardFocusManager.class) {
            this.checkKFMSecurity();
            return KeyboardFocusManager.focusOwner;
        }
    }
    
    protected void setGlobalFocusOwner(Component focusOwner) throws SecurityException {
        Component focusOwner2 = null;
        boolean n = false;
        if (focusOwner == null || focusOwner.isFocusable()) {
            synchronized (KeyboardFocusManager.class) {
                this.checkKFMSecurity();
                focusOwner2 = this.getFocusOwner();
                try {
                    this.fireVetoableChange("focusOwner", focusOwner2, focusOwner);
                }
                catch (PropertyVetoException ex) {
                    return;
                }
                KeyboardFocusManager.focusOwner = focusOwner;
                if (focusOwner != null && (this.getCurrentFocusCycleRoot() == null || !focusOwner.isFocusCycleRoot(this.getCurrentFocusCycleRoot()))) {
                    Container focusCycleRootAncestor = focusOwner.getFocusCycleRootAncestor();
                    if (focusCycleRootAncestor == null && focusOwner instanceof Window) {
                        focusCycleRootAncestor = (Container)focusOwner;
                    }
                    if (focusCycleRootAncestor != null) {
                        this.setGlobalCurrentFocusCycleRootPriv(focusCycleRootAncestor);
                    }
                }
                n = true;
            }
        }
        if (!(n == 0)) {
            this.firePropertyChange("focusOwner", focusOwner2, focusOwner);
        }
    }
    
    public void clearFocusOwner() {
        if (this.getFocusOwner() != null) {
            this.clearGlobalFocusOwner();
        }
    }
    
    public void clearGlobalFocusOwner() throws SecurityException {
        checkReplaceKFMPermission();
        if (!GraphicsEnvironment.isHeadless()) {
            Toolkit.getDefaultToolkit();
            this._clearGlobalFocusOwner();
        }
    }
    
    private void _clearGlobalFocusOwner() {
        this.peer.clearGlobalFocusOwner(markClearGlobalFocusOwner());
    }
    
    void clearGlobalFocusOwnerPriv() {
        AccessController.doPrivileged(new PrivilegedAction<Void>() {
            @Override
            public Void run() {
                KeyboardFocusManager.this.clearGlobalFocusOwner();
                return null;
            }
        });
    }
    
    Component getNativeFocusOwner() {
        return this.peer.getCurrentFocusOwner();
    }
    
    void setNativeFocusOwner(Component component) {
        if (KeyboardFocusManager.focusLog.isLoggable(PlatformLogger.Level.FINEST)) {
            KeyboardFocusManager.focusLog.finest("Calling peer {0} setCurrentFocusOwner for {1}", new Object[] { String.valueOf(this.peer), String.valueOf(component) });
        }
        this.peer.setCurrentFocusOwner(component);
    }
    
    Window getNativeFocusedWindow() {
        return this.peer.getCurrentFocusedWindow();
    }
    
    public Component getPermanentFocusOwner() {
        synchronized (KeyboardFocusManager.class) {
            if (KeyboardFocusManager.permanentFocusOwner == null) {
                return null;
            }
            return (KeyboardFocusManager.permanentFocusOwner.appContext == AppContext.getAppContext()) ? KeyboardFocusManager.permanentFocusOwner : null;
        }
    }
    
    protected Component getGlobalPermanentFocusOwner() throws SecurityException {
        synchronized (KeyboardFocusManager.class) {
            this.checkKFMSecurity();
            return KeyboardFocusManager.permanentFocusOwner;
        }
    }
    
    protected void setGlobalPermanentFocusOwner(Component permanentFocusOwner) throws SecurityException {
        Component permanentFocusOwner2 = null;
        boolean n = false;
        if (permanentFocusOwner == null || permanentFocusOwner.isFocusable()) {
            synchronized (KeyboardFocusManager.class) {
                this.checkKFMSecurity();
                permanentFocusOwner2 = this.getPermanentFocusOwner();
                try {
                    this.fireVetoableChange("permanentFocusOwner", permanentFocusOwner2, permanentFocusOwner);
                }
                catch (PropertyVetoException ex) {
                    return;
                }
                setMostRecentFocusOwner(KeyboardFocusManager.permanentFocusOwner = permanentFocusOwner);
                n = true;
            }
        }
        if (!(n == 0)) {
            this.firePropertyChange("permanentFocusOwner", permanentFocusOwner2, permanentFocusOwner);
        }
    }
    
    public Window getFocusedWindow() {
        synchronized (KeyboardFocusManager.class) {
            if (KeyboardFocusManager.focusedWindow == null) {
                return null;
            }
            return (KeyboardFocusManager.focusedWindow.appContext == AppContext.getAppContext()) ? KeyboardFocusManager.focusedWindow : null;
        }
    }
    
    protected Window getGlobalFocusedWindow() throws SecurityException {
        synchronized (KeyboardFocusManager.class) {
            this.checkKFMSecurity();
            return KeyboardFocusManager.focusedWindow;
        }
    }
    
    protected void setGlobalFocusedWindow(Window focusedWindow) throws SecurityException {
        Window focusedWindow2 = null;
        boolean n = false;
        if (focusedWindow == null || focusedWindow.isFocusableWindow()) {
            synchronized (KeyboardFocusManager.class) {
                this.checkKFMSecurity();
                focusedWindow2 = this.getFocusedWindow();
                try {
                    this.fireVetoableChange("focusedWindow", focusedWindow2, focusedWindow);
                }
                catch (PropertyVetoException ex) {
                    return;
                }
                KeyboardFocusManager.focusedWindow = focusedWindow;
                n = true;
            }
        }
        if (!(n == 0)) {
            this.firePropertyChange("focusedWindow", focusedWindow2, focusedWindow);
        }
    }
    
    public Window getActiveWindow() {
        synchronized (KeyboardFocusManager.class) {
            if (KeyboardFocusManager.activeWindow == null) {
                return null;
            }
            return (KeyboardFocusManager.activeWindow.appContext == AppContext.getAppContext()) ? KeyboardFocusManager.activeWindow : null;
        }
    }
    
    protected Window getGlobalActiveWindow() throws SecurityException {
        synchronized (KeyboardFocusManager.class) {
            this.checkKFMSecurity();
            return KeyboardFocusManager.activeWindow;
        }
    }
    
    protected void setGlobalActiveWindow(Window window) throws SecurityException {
        Window activeWindow;
        synchronized (KeyboardFocusManager.class) {
            this.checkKFMSecurity();
            activeWindow = this.getActiveWindow();
            if (KeyboardFocusManager.focusLog.isLoggable(PlatformLogger.Level.FINER)) {
                KeyboardFocusManager.focusLog.finer("Setting global active window to " + window + ", old active " + activeWindow);
            }
            try {
                this.fireVetoableChange("activeWindow", activeWindow, window);
            }
            catch (PropertyVetoException ex) {
                return;
            }
            KeyboardFocusManager.activeWindow = window;
        }
        this.firePropertyChange("activeWindow", activeWindow, window);
    }
    
    public synchronized FocusTraversalPolicy getDefaultFocusTraversalPolicy() {
        return this.defaultPolicy;
    }
    
    public void setDefaultFocusTraversalPolicy(FocusTraversalPolicy defaultPolicy) {
        if (defaultPolicy == null) {
            throw new IllegalArgumentException("default focus traversal policy cannot be null");
        }
        FocusTraversalPolicy defaultPolicy2;
        synchronized (this) {
            defaultPolicy2 = this.defaultPolicy;
            this.defaultPolicy = defaultPolicy;
        }
        this.firePropertyChange("defaultFocusTraversalPolicy", defaultPolicy2, defaultPolicy);
        return;
    }
    
    public void setDefaultFocusTraversalKeys(int n, Set<? extends AWTKeyStroke> c) {
        if (n < 0 || n >= 4) {
            throw new IllegalArgumentException("invalid focus traversal key identifier");
        }
        if (c == null) {
            throw new IllegalArgumentException("cannot set null Set of default focus traversal keys");
        }
        Set<Object> set;
        synchronized (this) {
            for (AWTKeyStroke awtKeyStroke : c) {
                if (awtKeyStroke == null) {
                    throw new IllegalArgumentException("cannot set null focus traversal key");
                }
                if (awtKeyStroke.getKeyChar() != '\uffff') {
                    throw new IllegalArgumentException("focus traversal keys cannot map to KEY_TYPED events");
                }
                for (int i = 0; i < 4; ++i) {
                    if (i != n && this.defaultFocusTraversalKeys[i].contains(awtKeyStroke)) {
                        throw new IllegalArgumentException("focus traversal keys must be unique for a Component");
                    }
                }
            }
            set = this.defaultFocusTraversalKeys[n];
            this.defaultFocusTraversalKeys[n] = Collections.unmodifiableSet(new HashSet(c));
        }
        this.firePropertyChange(KeyboardFocusManager.defaultFocusTraversalKeyPropertyNames[n], set, c);
        return;
    }
    
    public Set<AWTKeyStroke> getDefaultFocusTraversalKeys(int n) {
        if (n < 0 || n >= 4) {
            throw new IllegalArgumentException("invalid focus traversal key identifier");
        }
        return this.defaultFocusTraversalKeys[n];
    }
    
    public Container getCurrentFocusCycleRoot() {
        synchronized (KeyboardFocusManager.class) {
            if (KeyboardFocusManager.currentFocusCycleRoot == null) {
                return null;
            }
            return (KeyboardFocusManager.currentFocusCycleRoot.appContext == AppContext.getAppContext()) ? KeyboardFocusManager.currentFocusCycleRoot : null;
        }
    }
    
    protected Container getGlobalCurrentFocusCycleRoot() throws SecurityException {
        synchronized (KeyboardFocusManager.class) {
            this.checkKFMSecurity();
            return KeyboardFocusManager.currentFocusCycleRoot;
        }
    }
    
    public void setGlobalCurrentFocusCycleRoot(Container currentFocusCycleRoot) throws SecurityException {
        checkReplaceKFMPermission();
        Container currentFocusCycleRoot2;
        synchronized (KeyboardFocusManager.class) {
            currentFocusCycleRoot2 = this.getCurrentFocusCycleRoot();
            KeyboardFocusManager.currentFocusCycleRoot = currentFocusCycleRoot;
        }
        this.firePropertyChange("currentFocusCycleRoot", currentFocusCycleRoot2, currentFocusCycleRoot);
    }
    
    void setGlobalCurrentFocusCycleRootPriv(final Container container) {
        AccessController.doPrivileged(new PrivilegedAction<Void>() {
            @Override
            public Void run() {
                KeyboardFocusManager.this.setGlobalCurrentFocusCycleRoot(container);
                return null;
            }
        });
    }
    
    public void addPropertyChangeListener(PropertyChangeListener listener) {
        if (listener != null) {
            synchronized (this) {
                if (this.changeSupport == null) {
                    this.changeSupport = new PropertyChangeSupport(this);
                }
                this.changeSupport.addPropertyChangeListener(listener);
            }
        }
    }
    
    public void removePropertyChangeListener(PropertyChangeListener listener) {
        if (listener != null) {
            synchronized (this) {
                if (this.changeSupport != null) {
                    this.changeSupport.removePropertyChangeListener(listener);
                }
            }
        }
    }
    
    public synchronized PropertyChangeListener[] getPropertyChangeListeners() {
        if (this.changeSupport == null) {
            this.changeSupport = new PropertyChangeSupport(this);
        }
        return this.changeSupport.getPropertyChangeListeners();
    }
    
    public void addPropertyChangeListener(String propertyName, PropertyChangeListener listener) {
        if (listener != null) {
            synchronized (this) {
                if (this.changeSupport == null) {
                    this.changeSupport = new PropertyChangeSupport(this);
                }
                this.changeSupport.addPropertyChangeListener(propertyName, listener);
            }
        }
    }
    
    public void removePropertyChangeListener(String propertyName, PropertyChangeListener listener) {
        if (listener != null) {
            synchronized (this) {
                if (this.changeSupport != null) {
                    this.changeSupport.removePropertyChangeListener(propertyName, listener);
                }
            }
        }
    }
    
    public synchronized PropertyChangeListener[] getPropertyChangeListeners(String propertyName) {
        if (this.changeSupport == null) {
            this.changeSupport = new PropertyChangeSupport(this);
        }
        return this.changeSupport.getPropertyChangeListeners(propertyName);
    }
    
    protected void firePropertyChange(String propertyName, Object oldValue, Object newValue) {
        if (oldValue == newValue) {
            return;
        }
        PropertyChangeSupport changeSupport = this.changeSupport;
        if (changeSupport != null) {
            changeSupport.firePropertyChange(propertyName, oldValue, newValue);
        }
        return;
    }
    
    public void addVetoableChangeListener(VetoableChangeListener listener) {
        if (listener != null) {
            synchronized (this) {
                if (this.vetoableSupport == null) {
                    this.vetoableSupport = new VetoableChangeSupport(this);
                }
                this.vetoableSupport.addVetoableChangeListener(listener);
            }
        }
    }
    
    public void removeVetoableChangeListener(VetoableChangeListener listener) {
        if (listener != null) {
            synchronized (this) {
                if (this.vetoableSupport != null) {
                    this.vetoableSupport.removeVetoableChangeListener(listener);
                }
            }
        }
    }
    
    public synchronized VetoableChangeListener[] getVetoableChangeListeners() {
        if (this.vetoableSupport == null) {
            this.vetoableSupport = new VetoableChangeSupport(this);
        }
        return this.vetoableSupport.getVetoableChangeListeners();
    }
    
    public void addVetoableChangeListener(String propertyName, VetoableChangeListener listener) {
        if (listener != null) {
            synchronized (this) {
                if (this.vetoableSupport == null) {
                    this.vetoableSupport = new VetoableChangeSupport(this);
                }
                this.vetoableSupport.addVetoableChangeListener(propertyName, listener);
            }
        }
    }
    
    public void removeVetoableChangeListener(String propertyName, VetoableChangeListener listener) {
        if (listener != null) {
            synchronized (this) {
                if (this.vetoableSupport != null) {
                    this.vetoableSupport.removeVetoableChangeListener(propertyName, listener);
                }
            }
        }
    }
    
    public synchronized VetoableChangeListener[] getVetoableChangeListeners(String propertyName) {
        if (this.vetoableSupport == null) {
            this.vetoableSupport = new VetoableChangeSupport(this);
        }
        return this.vetoableSupport.getVetoableChangeListeners(propertyName);
    }
    
    protected void fireVetoableChange(String propertyName, Object oldValue, Object newValue) throws PropertyVetoException {
        if (oldValue == newValue) {
            return;
        }
        VetoableChangeSupport vetoableSupport = this.vetoableSupport;
        if (vetoableSupport != null) {
            vetoableSupport.fireVetoableChange(propertyName, oldValue, newValue);
        }
        return;
    }
    
    public void addKeyEventDispatcher(KeyEventDispatcher e) {
        if (e != null) {
            synchronized (this) {
                if (this.keyEventDispatchers == null) {
                    this.keyEventDispatchers = new LinkedList();
                }
                this.keyEventDispatchers.add(e);
            }
        }
    }
    
    public void removeKeyEventDispatcher(KeyEventDispatcher o) {
        if (o != null) {
            synchronized (this) {
                if (this.keyEventDispatchers != null) {
                    this.keyEventDispatchers.remove(o);
                }
            }
        }
    }
    
    protected synchronized List<KeyEventDispatcher> getKeyEventDispatchers() {
        return (this.keyEventDispatchers != null) ? ((List)this.keyEventDispatchers.clone()) : null;
    }
    
    public void addKeyEventPostProcessor(KeyEventPostProcessor e) {
        if (e != null) {
            synchronized (this) {
                if (this.keyEventPostProcessors == null) {
                    this.keyEventPostProcessors = new LinkedList();
                }
                this.keyEventPostProcessors.add(e);
            }
        }
    }
    
    public void removeKeyEventPostProcessor(KeyEventPostProcessor o) {
        if (o != null) {
            synchronized (this) {
                if (this.keyEventPostProcessors != null) {
                    this.keyEventPostProcessors.remove(o);
                }
            }
        }
    }
    
    protected List<KeyEventPostProcessor> getKeyEventPostProcessors() {
        return (this.keyEventPostProcessors != null) ? ((List)this.keyEventPostProcessors.clone()) : null;
    }
    
    static void setMostRecentFocusOwner(Component parent) {
        for (; parent != null && !(parent instanceof Window); parent = parent.parent) {}
        if (parent != null) {
            setMostRecentFocusOwner((Window)parent, parent);
        }
    }
    
    static synchronized void setMostRecentFocusOwner(Window window, Component referent) {
        WeakReference<Component> weakReference = null;
        if (referent != null) {
            weakReference = new WeakReference(referent);
        }
        KeyboardFocusManager.mostRecentFocusOwners.put(window, weakReference);
    }
    
    static void clearMostRecentFocusOwner(Component component) {
        if (component == null) {
            return;
        }
        Container container;
        synchronized (component.getTreeLock()) {
            for (container = component.getParent(); container != null && !(container instanceof Window); container = container.getParent()) {}
        }
        synchronized (KeyboardFocusManager.class) {
            if (container != null && getMostRecentFocusOwner((Window)container) == component) {
                setMostRecentFocusOwner((Window)container, null);
            }
            if (container != null) {
                Window window = (Window)container;
                if (window.getTemporaryLostComponent() == component) {
                    window.setTemporaryLostComponent(null);
                }
            }
        }
        return;
    }
    
    static synchronized Component getMostRecentFocusOwner(Window window) {
        WeakReference<Object> weakReference = (WeakReference)KeyboardFocusManager.mostRecentFocusOwners.get(window);
        return (weakReference == null) ? null : ((Component)weakReference.get());
    }
    
    public abstract boolean dispatchEvent(AWTEvent p0);
    
    public final void redispatchEvent(Component component, AWTEvent awtEvent) {
        awtEvent.focusManagerIsDispatching = true;
        component.dispatchEvent(awtEvent);
        awtEvent.focusManagerIsDispatching = false;
    }
    
    @Override
    public abstract boolean dispatchKeyEvent(KeyEvent p0);
    
    @Override
    public abstract boolean postProcessKeyEvent(KeyEvent p0);
    
    public abstract void processKeyEvent(Component p0, KeyEvent p1);
    
    protected abstract void enqueueKeyEvents(long p0, Component p1);
    
    protected abstract void dequeueKeyEvents(long p0, Component p1);
    
    protected abstract void discardKeyEvents(Component p0);
    
    public abstract void focusNextComponent(Component p0);
    
    public abstract void focusPreviousComponent(Component p0);
    
    public abstract void upFocusCycle(Component p0);
    
    public abstract void downFocusCycle(Container p0);
    
    public final void focusNextComponent() {
        Component focusOwner = this.getFocusOwner();
        if (focusOwner != null) {
            this.focusNextComponent(focusOwner);
        }
    }
    
    public final void focusPreviousComponent() {
        Component focusOwner = this.getFocusOwner();
        if (focusOwner != null) {
            this.focusPreviousComponent(focusOwner);
        }
    }
    
    public final void upFocusCycle() {
        Component focusOwner = this.getFocusOwner();
        if (focusOwner != null) {
            this.upFocusCycle(focusOwner);
        }
    }
    
    public final void downFocusCycle() {
        Component focusOwner = this.getFocusOwner();
        if ((focusOwner instanceof Container)) {
            this.downFocusCycle((Container)focusOwner);
        }
    }
    
    void dumpRequests() {
        System.err.println(">>> Requests dump, time: " + System.currentTimeMillis());
        synchronized (KeyboardFocusManager.heavyweightRequests) {
            Iterator<Object> iterator = KeyboardFocusManager.heavyweightRequests.iterator();
            while (iterator.hasNext()) {
                System.err.println(">>> Req: " + (Object)iterator.next());
            }
        }
        System.err.println("");
    }
    
    static boolean processSynchronousLightweightTransfer(Component comp, Component source, boolean b, boolean b2, long n) {
        Window containingWindow = SunToolkit.getContainingWindow(comp);
        if (containingWindow == null || !containingWindow.syncLWRequests) {
            return false;
        }
        if (source == null) {
            source = comp;
        }
        KeyboardFocusManager currentKeyboardFocusManager = getCurrentKeyboardFocusManager(SunToolkit.targetToAppContext(source));
        Object o = null;
        Object o2 = null;
        Component globalFocusOwner = currentKeyboardFocusManager.getGlobalFocusOwner();
        synchronized (KeyboardFocusManager.heavyweightRequests) {
            if (getLastHWRequest() == null && comp == currentKeyboardFocusManager.getNativeFocusOwner() && KeyboardFocusManager.allowSyncFocusRequests) {
                if (source == globalFocusOwner) {
                    return true;
                }
                currentKeyboardFocusManager.enqueueKeyEvents(n, source);
                KeyboardFocusManager.heavyweightRequests.add(new KeyboardFocusManager.HeavyweightFocusRequest(comp, source, b, CausedFocusEvent.Cause.UNKNOWN));
                if (globalFocusOwner != null) {
                    o = new FocusEvent(globalFocusOwner, 1005, b, source);
                }
                o2 = new FocusEvent(source, 1004, b, globalFocusOwner);
            }
        }
        boolean n2 = false;
        int clearingCurrentLightweightRequests = KeyboardFocusManager.clearingCurrentLightweightRequests;
        Throwable t = null;
        try {
            KeyboardFocusManager.clearingCurrentLightweightRequests = false;
            synchronized (Component.LOCK) {
                if (o != null && globalFocusOwner != null) {
                    o.isPosted = true;
                    t = dispatchAndCatchException(t, globalFocusOwner, o);
                    n2 = true;
                }
                if (o2 != null && source != null) {
                    o2.isPosted = true;
                    t = dispatchAndCatchException(t, source, o2);
                    n2 = true;
                }
            }
        }
        finally {
            KeyboardFocusManager.clearingCurrentLightweightRequests = clearingCurrentLightweightRequests;
        }
        if ((t instanceof RuntimeException)) {
            throw (RuntimeException)t;
        }
        if ((t instanceof Error)) {
            throw (Error)t;
        }
        return n2;
    }
    
    static int shouldNativelyFocusHeavyweight(Component component, Component obj, boolean b, boolean b2, long n, CausedFocusEvent.Cause cause) {
        if (KeyboardFocusManager.log.isLoggable(PlatformLogger.Level.FINE)) {
            if (component == null) {
                KeyboardFocusManager.log.fine("Assertion (heavyweight != null) failed");
            }
            if (n == 0L) {
                KeyboardFocusManager.log.fine("Assertion (time != 0) failed");
            }
        }
        if (obj == null) {
            obj = component;
        }
        KeyboardFocusManager currentKeyboardFocusManager = getCurrentKeyboardFocusManager(SunToolkit.targetToAppContext(obj));
        KeyboardFocusManager currentKeyboardFocusManager2 = getCurrentKeyboardFocusManager();
        Component globalFocusOwner = currentKeyboardFocusManager2.getGlobalFocusOwner();
        Component nativeFocusOwner = currentKeyboardFocusManager2.getNativeFocusOwner();
        Window nativeFocusedWindow = currentKeyboardFocusManager2.getNativeFocusedWindow();
        if (KeyboardFocusManager.focusLog.isLoggable(PlatformLogger.Level.FINER)) {
            KeyboardFocusManager.focusLog.finer("SNFH for {0} in {1}", new Object[] { String.valueOf(obj), String.valueOf(component) });
        }
        if (KeyboardFocusManager.focusLog.isLoggable(PlatformLogger.Level.FINEST)) {
            KeyboardFocusManager.focusLog.finest("0. Current focus owner {0}", new Object[] { String.valueOf(globalFocusOwner) });
            KeyboardFocusManager.focusLog.finest("0. Native focus owner {0}", new Object[] { String.valueOf(nativeFocusOwner) });
            KeyboardFocusManager.focusLog.finest("0. Native focused window {0}", new Object[] { String.valueOf(nativeFocusedWindow) });
        }
        synchronized (KeyboardFocusManager.heavyweightRequests) {
            KeyboardFocusManager.HeavyweightFocusRequest lastHWRequest = getLastHWRequest();
            if (KeyboardFocusManager.focusLog.isLoggable(PlatformLogger.Level.FINEST)) {
                KeyboardFocusManager.focusLog.finest("Request {0}", new Object[] { String.valueOf(lastHWRequest) });
            }
            if (lastHWRequest == null && component == nativeFocusOwner && component.getContainingWindow() == nativeFocusedWindow) {
                if (obj == globalFocusOwner) {
                    if (KeyboardFocusManager.focusLog.isLoggable(PlatformLogger.Level.FINEST)) {
                        KeyboardFocusManager.focusLog.finest("1. SNFH_FAILURE for {0}", new Object[] { String.valueOf(obj) });
                    }
                    return 0;
                }
                else {
                    currentKeyboardFocusManager.enqueueKeyEvents(n, obj);
                    KeyboardFocusManager.heavyweightRequests.add(new KeyboardFocusManager.HeavyweightFocusRequest(component, obj, b, cause));
                    if (globalFocusOwner != null) {
                        SunToolkit.postEvent(globalFocusOwner.appContext, new CausedFocusEvent(globalFocusOwner, 1005, b, obj, cause));
                    }
                    SunToolkit.postEvent(obj.appContext, new CausedFocusEvent(obj, 1004, b, globalFocusOwner, cause));
                    if (KeyboardFocusManager.focusLog.isLoggable(PlatformLogger.Level.FINEST)) {
                        KeyboardFocusManager.focusLog.finest("2. SNFH_HANDLED for {0}", new Object[] { String.valueOf(obj) });
                    }
                    return 1;
                }
            }
            else if (lastHWRequest != null && lastHWRequest.heavyweight == component) {
                if (lastHWRequest.addLightweightRequest(obj, b, cause)) {
                    currentKeyboardFocusManager.enqueueKeyEvents(n, obj);
                }
                if (KeyboardFocusManager.focusLog.isLoggable(PlatformLogger.Level.FINEST)) {
                    KeyboardFocusManager.focusLog.finest("3. SNFH_HANDLED for lightweight" + obj + " in " + component);
                }
                return 1;
            }
            else {
                if (!b2) {
                    if (lastHWRequest == KeyboardFocusManager.HeavyweightFocusRequest.CLEAR_GLOBAL_FOCUS_OWNER) {
                        int size = KeyboardFocusManager.heavyweightRequests.size();
                        lastHWRequest = ((size >= 2) ? ((KeyboardFocusManager.HeavyweightFocusRequest)KeyboardFocusManager.heavyweightRequests.get(size - 2)) : null);
                    }
                    if (focusedWindowChanged(component, (lastHWRequest != null) ? lastHWRequest.heavyweight : nativeFocusedWindow)) {
                        if (KeyboardFocusManager.focusLog.isLoggable(PlatformLogger.Level.FINEST)) {
                            KeyboardFocusManager.focusLog.finest("4. SNFH_FAILURE for " + obj);
                        }
                        return 0;
                    }
                }
                currentKeyboardFocusManager.enqueueKeyEvents(n, obj);
                KeyboardFocusManager.heavyweightRequests.add(new KeyboardFocusManager.HeavyweightFocusRequest(component, obj, b, cause));
                if (KeyboardFocusManager.focusLog.isLoggable(PlatformLogger.Level.FINEST)) {
                    KeyboardFocusManager.focusLog.finest("5. SNFH_PROCEED for " + obj);
                }
                return 2;
            }
        }
    }
    
    static Window markClearGlobalFocusOwner() {
        Window window = getCurrentKeyboardFocusManager().getNativeFocusedWindow();
        synchronized (KeyboardFocusManager.heavyweightRequests) {
            KeyboardFocusManager.HeavyweightFocusRequest lastHWRequest = getLastHWRequest();
            if (lastHWRequest == KeyboardFocusManager.HeavyweightFocusRequest.CLEAR_GLOBAL_FOCUS_OWNER) {
                return null;
            }
            KeyboardFocusManager.heavyweightRequests.add(KeyboardFocusManager.HeavyweightFocusRequest.CLEAR_GLOBAL_FOCUS_OWNER);
            for (window = ((lastHWRequest != null) ? SunToolkit.getContainingWindow(lastHWRequest.heavyweight) : window); window != null && !(window instanceof Frame) && !(window instanceof Dialog); window = window.getParent_NoClientCode()) {}
            return (Window)window;
        }
    }
    
    Component getCurrentWaitingRequest(Component component) {
        synchronized (KeyboardFocusManager.heavyweightRequests) {
            KeyboardFocusManager.HeavyweightFocusRequest firstHWRequest = getFirstHWRequest();
            if (firstHWRequest != null && firstHWRequest.heavyweight == component) {
                KeyboardFocusManager.LightweightFocusRequest lightweightFocusRequest = (KeyboardFocusManager.LightweightFocusRequest)firstHWRequest.lightweightRequests.getFirst();
                if (lightweightFocusRequest != null) {
                    return lightweightFocusRequest.component;
                }
            }
        }
        return null;
    }
    
    static boolean isAutoFocusTransferEnabled() {
        synchronized (KeyboardFocusManager.heavyweightRequests) {
            return KeyboardFocusManager.heavyweightRequests.size() == 0 && !KeyboardFocusManager.disableRestoreFocus && null == KeyboardFocusManager.currentLightweightRequests;
        }
    }
    
    static boolean isAutoFocusTransferEnabledFor(Component component) {
        return isAutoFocusTransferEnabled() && component.isAutoFocusTransferOnDisposal();
    }
    
    private static Throwable dispatchAndCatchException(Throwable t, Component component, FocusEvent focusEvent) {
        Throwable t2 = null;
        try {
            component.dispatchEvent(focusEvent);
        }
        catch (RuntimeException ex) {
            t2 = ex;
        }
        catch (Error error) {
            t2 = error;
        }
        if (t2 != null) {
            if (t != null) {
                handleException(t);
                return t2;
            }
            return t2;
        }
        else {
            return t;
        }
    }
    
    private static void handleException(Throwable t) {
        t.printStackTrace();
    }
    
    static void processCurrentLightweightRequests() {
        KeyboardFocusManager currentKeyboardFocusManager = getCurrentKeyboardFocusManager();
        LinkedList<Object> currentLightweightRequests = null;
        Component globalFocusOwner = currentKeyboardFocusManager.getGlobalFocusOwner();
        if (globalFocusOwner != null && globalFocusOwner.appContext != AppContext.getAppContext()) {
            return;
        }
        synchronized (KeyboardFocusManager.heavyweightRequests) {
            if (KeyboardFocusManager.currentLightweightRequests != null) {
                KeyboardFocusManager.clearingCurrentLightweightRequests = true;
                KeyboardFocusManager.disableRestoreFocus = true;
                currentLightweightRequests = KeyboardFocusManager.currentLightweightRequests;
                KeyboardFocusManager.allowSyncFocusRequests = (currentLightweightRequests.size() < 2);
                KeyboardFocusManager.currentLightweightRequests = null;
            }
            else {
                return;
            }
        }
        Throwable t = null;
        try {
            if (currentLightweightRequests != null) {
                Component component = null;
                Iterator<Object> iterator = currentLightweightRequests.iterator();
                while (iterator.hasNext()) {
                    Component globalFocusOwner2 = currentKeyboardFocusManager.getGlobalFocusOwner();
                    KeyboardFocusManager.LightweightFocusRequest lightweightFocusRequest = (KeyboardFocusManager.LightweightFocusRequest)iterator.next();
                    if (!iterator.hasNext()) {
                        KeyboardFocusManager.disableRestoreFocus = false;
                    }
                    Object o = null;
                    if (globalFocusOwner2 != null) {
                        o = new CausedFocusEvent(globalFocusOwner2, 1005, lightweightFocusRequest.temporary, lightweightFocusRequest.component, lightweightFocusRequest.cause);
                    }
                    CausedFocusEvent o2 = new CausedFocusEvent(lightweightFocusRequest.component, 1004, lightweightFocusRequest.temporary, (globalFocusOwner2 == null) ? component : globalFocusOwner2, lightweightFocusRequest.cause);
                    if (globalFocusOwner2 != null) {
                        o.isPosted = true;
                        t = dispatchAndCatchException(t, globalFocusOwner2, o);
                    }
                    o2.isPosted = true;
                    t = dispatchAndCatchException(t, lightweightFocusRequest.component, o2);
                    if (currentKeyboardFocusManager.getGlobalFocusOwner() == lightweightFocusRequest.component) {
                        component = lightweightFocusRequest.component;
                    }
                    else {
                        continue;
                    }
                }
            }
        }
        finally {
            KeyboardFocusManager.clearingCurrentLightweightRequests = false;
            KeyboardFocusManager.disableRestoreFocus = false;
            KeyboardFocusManager.allowSyncFocusRequests = true;
        }
        if ((t instanceof RuntimeException)) {
            throw (RuntimeException)t;
        }
        if ((t instanceof Error)) {
            throw (Error)t;
        }
        return;
    }
    
    static FocusEvent retargetUnexpectedFocusEvent(FocusEvent focusEvent) {
        synchronized (KeyboardFocusManager.heavyweightRequests) {
            if (removeFirstRequest()) {
                return (FocusEvent)retargetFocusEvent(focusEvent);
            }
            Component component = focusEvent.getComponent();
            Component oppositeComponent = focusEvent.getOppositeComponent();
            boolean temporary = false;
            if (focusEvent.getID() == 1005 && (oppositeComponent == null || isTemporary(oppositeComponent, component))) {
                temporary = true;
            }
            return new CausedFocusEvent(component, focusEvent.getID(), (boolean)temporary, oppositeComponent, CausedFocusEvent.Cause.NATIVE_SYSTEM);
        }
    }
    
    static FocusEvent retargetFocusGained(FocusEvent focusEvent) {
        assert focusEvent.getID() == 1004;
        Component globalFocusOwner = getCurrentKeyboardFocusManager().getGlobalFocusOwner();
        Component component = focusEvent.getComponent();
        Component oppositeComponent = focusEvent.getOppositeComponent();
        Component component2 = getHeavyweight(component);
        synchronized (KeyboardFocusManager.heavyweightRequests) {
            KeyboardFocusManager.HeavyweightFocusRequest firstHWRequest = getFirstHWRequest();
            if (firstHWRequest == KeyboardFocusManager.HeavyweightFocusRequest.CLEAR_GLOBAL_FOCUS_OWNER) {
                return retargetUnexpectedFocusEvent(focusEvent);
            }
            if (component != null && component2 == null && firstHWRequest != null && component == firstHWRequest.getFirstLightweightRequest().component) {
                component = (component2 = firstHWRequest.heavyweight);
            }
            if (firstHWRequest != null && component2 == firstHWRequest.heavyweight) {
                KeyboardFocusManager.heavyweightRequests.removeFirst();
                KeyboardFocusManager.LightweightFocusRequest lightweightFocusRequest = (KeyboardFocusManager.LightweightFocusRequest)firstHWRequest.lightweightRequests.removeFirst();
                Component component3 = lightweightFocusRequest.component;
                if (globalFocusOwner != null) {
                    KeyboardFocusManager.newFocusOwner = component3;
                }
                int temporary = oppositeComponent != null && !isTemporary(component3, oppositeComponent) && lightweightFocusRequest.temporary;
                if (firstHWRequest.lightweightRequests.size() > 0) {
                    KeyboardFocusManager.currentLightweightRequests = firstHWRequest.lightweightRequests;
                    EventQueue.invokeLater(new Runnable() {
                        @Override
                        public void run() {
                            KeyboardFocusManager.processCurrentLightweightRequests();
                        }
                    });
                }
                return new CausedFocusEvent(component3, 1004, (boolean)temporary, oppositeComponent, lightweightFocusRequest.cause);
            }
            else {
                if (globalFocusOwner != null && globalFocusOwner.getContainingWindow() == component && (firstHWRequest == null || component != firstHWRequest.heavyweight)) {
                    return new CausedFocusEvent(globalFocusOwner, 1004, false, null, CausedFocusEvent.Cause.ACTIVATION);
                }
                return retargetUnexpectedFocusEvent(focusEvent);
            }
        }
    }
    
    static FocusEvent retargetFocusLost(FocusEvent focusEvent) {
        assert focusEvent.getID() == 1005;
        Component globalFocusOwner = getCurrentKeyboardFocusManager().getGlobalFocusOwner();
        Component oppositeComponent = focusEvent.getOppositeComponent();
        Component heavyweight = getHeavyweight(oppositeComponent);
        synchronized (KeyboardFocusManager.heavyweightRequests) {
            KeyboardFocusManager.HeavyweightFocusRequest firstHWRequest = getFirstHWRequest();
            if (firstHWRequest == KeyboardFocusManager.HeavyweightFocusRequest.CLEAR_GLOBAL_FOCUS_OWNER) {
                if (globalFocusOwner != null) {
                    KeyboardFocusManager.heavyweightRequests.removeFirst();
                    return new CausedFocusEvent(globalFocusOwner, 1005, false, null, CausedFocusEvent.Cause.CLEAR_GLOBAL_FOCUS_OWNER);
                }
            }
            else if (oppositeComponent == null) {
                if (globalFocusOwner != null) {
                    return new CausedFocusEvent(globalFocusOwner, 1005, true, null, CausedFocusEvent.Cause.ACTIVATION);
                }
                return focusEvent;
            }
            else if (firstHWRequest != null && (heavyweight == firstHWRequest.heavyweight || (heavyweight == null && oppositeComponent == firstHWRequest.getFirstLightweightRequest().component))) {
                if (globalFocusOwner == null) {
                    return focusEvent;
                }
                KeyboardFocusManager.LightweightFocusRequest lightweightFocusRequest = (KeyboardFocusManager.LightweightFocusRequest)firstHWRequest.lightweightRequests.getFirst();
                return new CausedFocusEvent(globalFocusOwner, 1005, isTemporary(oppositeComponent, globalFocusOwner) || lightweightFocusRequest.temporary, lightweightFocusRequest.component, lightweightFocusRequest.cause);
            }
            else if (focusedWindowChanged(oppositeComponent, globalFocusOwner)) {
                if (!focusEvent.isTemporary() && globalFocusOwner != null) {
                    focusEvent = new CausedFocusEvent(globalFocusOwner, 1005, true, oppositeComponent, CausedFocusEvent.Cause.ACTIVATION);
                }
                return focusEvent;
            }
            return retargetUnexpectedFocusEvent(focusEvent);
        }
    }
    
    static AWTEvent retargetFocusEvent(AWTEvent awtEvent) {
        if (KeyboardFocusManager.clearingCurrentLightweightRequests) {
            return awtEvent;
        }
        KeyboardFocusManager currentKeyboardFocusManager = getCurrentKeyboardFocusManager();
        if (KeyboardFocusManager.focusLog.isLoggable(PlatformLogger.Level.FINER)) {
            if (awtEvent instanceof FocusEvent || awtEvent instanceof WindowEvent) {
                KeyboardFocusManager.focusLog.finer(">>> {0}", new Object[] { String.valueOf(awtEvent) });
            }
            if (KeyboardFocusManager.focusLog.isLoggable(PlatformLogger.Level.FINER) && awtEvent instanceof KeyEvent) {
                KeyboardFocusManager.focusLog.finer("    focus owner is {0}", new Object[] { String.valueOf(currentKeyboardFocusManager.getGlobalFocusOwner()) });
                KeyboardFocusManager.focusLog.finer(">>> {0}", new Object[] { String.valueOf(awtEvent) });
            }
        }
        synchronized (KeyboardFocusManager.heavyweightRequests) {
            if (KeyboardFocusManager.newFocusOwner != null && awtEvent.getID() == 1005) {
                FocusEvent focusEvent = (FocusEvent)awtEvent;
                if (currentKeyboardFocusManager.getGlobalFocusOwner() == focusEvent.getComponent() && focusEvent.getOppositeComponent() == KeyboardFocusManager.newFocusOwner) {
                    KeyboardFocusManager.newFocusOwner = null;
                    return awtEvent;
                }
            }
        }
        processCurrentLightweightRequests();
        switch (awtEvent.getID()) {
            case 1004: {
                awtEvent = retargetFocusGained((FocusEvent)awtEvent);
                break;
            }
            case 1005: {
                awtEvent = retargetFocusLost((FocusEvent)awtEvent);
                break;
            }
        }
        return awtEvent;
    }
    
    void clearMarkers() {
    }
    
    static boolean removeFirstRequest() {
        KeyboardFocusManager currentKeyboardFocusManager = getCurrentKeyboardFocusManager();
        synchronized (KeyboardFocusManager.heavyweightRequests) {
            KeyboardFocusManager.HeavyweightFocusRequest firstHWRequest = getFirstHWRequest();
            if (firstHWRequest != null) {
                KeyboardFocusManager.heavyweightRequests.removeFirst();
                if (firstHWRequest.lightweightRequests != null) {
                    Iterator<Object> iterator = firstHWRequest.lightweightRequests.iterator();
                    while (iterator.hasNext()) {
                        currentKeyboardFocusManager.dequeueKeyEvents(-1L, ((KeyboardFocusManager.LightweightFocusRequest)iterator.next()).component);
                    }
                }
            }
            if (KeyboardFocusManager.heavyweightRequests.size() == 0) {
                currentKeyboardFocusManager.clearMarkers();
            }
            return KeyboardFocusManager.heavyweightRequests.size() > 0;
        }
    }
    
    static void removeLastFocusRequest(Component component) {
        if (KeyboardFocusManager.log.isLoggable(PlatformLogger.Level.FINE) && component == null) {
            KeyboardFocusManager.log.fine("Assertion (heavyweight != null) failed");
        }
        KeyboardFocusManager currentKeyboardFocusManager = getCurrentKeyboardFocusManager();
        synchronized (KeyboardFocusManager.heavyweightRequests) {
            KeyboardFocusManager.HeavyweightFocusRequest lastHWRequest = getLastHWRequest();
            if (lastHWRequest != null && lastHWRequest.heavyweight == component) {
                KeyboardFocusManager.heavyweightRequests.removeLast();
            }
            if (KeyboardFocusManager.heavyweightRequests.size() == 0) {
                currentKeyboardFocusManager.clearMarkers();
            }
        }
    }
    
    private static boolean focusedWindowChanged(Component comp, Component comp2) {
        Window containingWindow = SunToolkit.getContainingWindow(comp);
        Window containingWindow2 = SunToolkit.getContainingWindow(comp2);
        return (containingWindow == null && containingWindow2 == null) || containingWindow == null || containingWindow2 == null || containingWindow != containingWindow2;
    }
    
    private static boolean isTemporary(Component comp, Component comp2) {
        Window containingWindow = SunToolkit.getContainingWindow(comp);
        Window containingWindow2 = SunToolkit.getContainingWindow(comp2);
        return (containingWindow != null || containingWindow2 != null) && (containingWindow == null || (containingWindow2 != null && containingWindow != containingWindow2));
    }
    
    static Component getHeavyweight(Component component) {
        if (component == null || component.getPeer() == null) {
            return null;
        }
        if ((component.getPeer() instanceof LightweightPeer)) {
            return component.getNativeContainer();
        }
        return component;
    }
    
    private static boolean isProxyActiveImpl(KeyEvent obj) {
        if (KeyboardFocusManager.proxyActive == null) {
            KeyboardFocusManager.proxyActive = (Field)AccessController.doPrivileged(new PrivilegedAction<Field>() {
                @Override
                public Field run() {
                    Field declaredField = null;
                    try {
                        declaredField = KeyEvent.class.getDeclaredField("isProxyActive");
                        if (declaredField != null) {
                            declaredField.setAccessible(true);
                        }
                    }
                    catch (NoSuchFieldException ex) {
                        assert false;
                    }
                    return declaredField;
                }
            });
        }
        try {
            return KeyboardFocusManager.proxyActive.getBoolean(obj);
        }
        catch (IllegalAccessException ex) {
            assert false;
            return false;
        }
    }
    
    static boolean isProxyActive(KeyEvent keyEvent) {
        return !GraphicsEnvironment.isHeadless() && isProxyActiveImpl(keyEvent);
    }
    
    private static KeyboardFocusManager.HeavyweightFocusRequest getLastHWRequest() {
        synchronized (KeyboardFocusManager.heavyweightRequests) {
            return (KeyboardFocusManager.heavyweightRequests.size() > 0) ? ((KeyboardFocusManager.HeavyweightFocusRequest)KeyboardFocusManager.heavyweightRequests.getLast()) : null;
        }
    }
    
    private static KeyboardFocusManager.HeavyweightFocusRequest getFirstHWRequest() {
        synchronized (KeyboardFocusManager.heavyweightRequests) {
            return (KeyboardFocusManager.heavyweightRequests.size() > 0) ? ((KeyboardFocusManager.HeavyweightFocusRequest)KeyboardFocusManager.heavyweightRequests.getFirst()) : null;
        }
    }
    
    private static void checkReplaceKFMPermission() throws SecurityException {
        SecurityManager securityManager = System.getSecurityManager();
        if (securityManager != null) {
            if (KeyboardFocusManager.replaceKeyboardFocusManagerPermission == null) {
                KeyboardFocusManager.replaceKeyboardFocusManagerPermission = new AWTPermission("replaceKeyboardFocusManager");
            }
            securityManager.checkPermission(KeyboardFocusManager.replaceKeyboardFocusManagerPermission);
        }
    }
    
    private void checkKFMSecurity() throws SecurityException {
        if (this != getCurrentKeyboardFocusManager()) {
            checkReplaceKFMPermission();
        }
    }
    
    static {
        focusLog = PlatformLogger.getLogger("java.awt.focus.KeyboardFocusManager");
        Toolkit.loadLibraries();
        if (!GraphicsEnvironment.isHeadless()) {
            initIDs();
        }
        AWTAccessor.setKeyboardFocusManagerAccessor(new AWTAccessor.KeyboardFocusManagerAccessor() {
            public int shouldNativelyFocusHeavyweight(Component component, Component component2, boolean b, boolean b2, long n, CausedFocusEvent.Cause cause) {
                return KeyboardFocusManager.shouldNativelyFocusHeavyweight(component, component2, b, b2, n, cause);
            }
            
            @Override
            public boolean processSynchronousLightweightTransfer(Component component, Component component2, boolean b, boolean b2, long n) {
                return KeyboardFocusManager.processSynchronousLightweightTransfer(component, component2, b, b2, n);
            }
            
            @Override
            public void removeLastFocusRequest(Component component) {
                KeyboardFocusManager.removeLastFocusRequest(component);
            }
            
            @Override
            public void setMostRecentFocusOwner(Window window, Component component) {
                KeyboardFocusManager.setMostRecentFocusOwner(window, component);
            }
            
            @Override
            public KeyboardFocusManager getCurrentKeyboardFocusManager(AppContext appContext) {
                return KeyboardFocusManager.getCurrentKeyboardFocusManager(appContext);
            }
            
            @Override
            public Container getCurrentFocusCycleRoot() {
                return KeyboardFocusManager.currentFocusCycleRoot;
            }
        });
        log = PlatformLogger.getLogger("java.awt.KeyboardFocusManager");
        defaultFocusTraversalKeyPropertyNames = new String[] { "forwardDefaultFocusTraversalKeys", "backwardDefaultFocusTraversalKeys", "upCycleDefaultFocusTraversalKeys", "downCycleDefaultFocusTraversalKeys" };
        defaultFocusTraversalKeyStrokes = new AWTKeyStroke[][] { { AWTKeyStroke.getAWTKeyStroke(9, 0, false), AWTKeyStroke.getAWTKeyStroke(9, 130, false) }, { AWTKeyStroke.getAWTKeyStroke(9, 65, false), AWTKeyStroke.getAWTKeyStroke(9, 195, false) }, new AWTKeyStroke[0], new AWTKeyStroke[0] };
        KeyboardFocusManager.mostRecentFocusOwners = new WeakHashMap();
        KeyboardFocusManager.heavyweightRequests = new LinkedList();
        KeyboardFocusManager.allowSyncFocusRequests = true;
        KeyboardFocusManager.newFocusOwner = null;
    }
    
    private static final class LightweightFocusRequest
    {
        final Component component;
        final boolean temporary;
        final CausedFocusEvent.Cause cause;
        
        LightweightFocusRequest(Component component, boolean temporary, CausedFocusEvent.Cause cause) {
            this.component = component;
            this.temporary = temporary;
            this.cause = cause;
        }
        
        @Override
        public String toString() {
            return "LightweightFocusRequest[component=" + this.component + ",temporary=" + this.temporary + ", cause=" + this.cause + "]";
        }
    }
    
    private static final class HeavyweightFocusRequest
    {
        final Component heavyweight;
        final LinkedList<KeyboardFocusManager.LightweightFocusRequest> lightweightRequests;
        static final KeyboardFocusManager.HeavyweightFocusRequest CLEAR_GLOBAL_FOCUS_OWNER;
        
        private HeavyweightFocusRequest() {
            this.heavyweight = null;
            this.lightweightRequests = null;
        }
        
        HeavyweightFocusRequest(Component heavyweight, Component component, boolean b, CausedFocusEvent.Cause cause) {
            if (KeyboardFocusManager.log.isLoggable(PlatformLogger.Level.FINE) && heavyweight == null) {
                KeyboardFocusManager.log.fine("Assertion (heavyweight != null) failed");
            }
            this.heavyweight = heavyweight;
            this.lightweightRequests = new LinkedList();
            this.addLightweightRequest(component, b, cause);
        }
        
        boolean addLightweightRequest(Component component, boolean b, CausedFocusEvent.Cause cause) {
            if (KeyboardFocusManager.log.isLoggable(PlatformLogger.Level.FINE)) {
                if (this == KeyboardFocusManager.HeavyweightFocusRequest.CLEAR_GLOBAL_FOCUS_OWNER) {
                    KeyboardFocusManager.log.fine("Assertion (this != HeavyweightFocusRequest.CLEAR_GLOBAL_FOCUS_OWNER) failed");
                }
                if (component == null) {
                    KeyboardFocusManager.log.fine("Assertion (descendant != null) failed");
                }
            }
            if (component != ((this.lightweightRequests.size() > 0) ? ((KeyboardFocusManager.LightweightFocusRequest)this.lightweightRequests.getLast()).component : null)) {
                this.lightweightRequests.add(new KeyboardFocusManager.LightweightFocusRequest(component, b, cause));
                return true;
            }
            return false;
        }
        
        KeyboardFocusManager.LightweightFocusRequest getFirstLightweightRequest() {
            if (this == KeyboardFocusManager.HeavyweightFocusRequest.CLEAR_GLOBAL_FOCUS_OWNER) {
                return null;
            }
            return (KeyboardFocusManager.LightweightFocusRequest)this.lightweightRequests.getFirst();
        }
        
        @Override
        public String toString() {
            int n = 1;
            String string = "HeavyweightFocusRequest[heavweight=" + this.heavyweight + ",lightweightRequests=";
            String str2;
            if (this.lightweightRequests == null) {
                str2 = string + null;
            }
            else {
                String str = string + "[";
                for (KeyboardFocusManager.LightweightFocusRequest obj : this.lightweightRequests) {
                    if (n != 0) {
                        n = 0;
                    }
                    else {
                        str += ",";
                    }
                    str += obj;
                }
                str2 = str + "]";
            }
            return str2 + "]";
        }
        
        static {
            CLEAR_GLOBAL_FOCUS_OWNER = new KeyboardFocusManager.HeavyweightFocusRequest();
        }
    }
}
